﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace PWTEditor.SaveFile
{
    public abstract class ChecksumBlock
    {
        private byte[] data;
        public virtual byte[] Data
        {
            get
            {
                return data;
            }
            set
            {
                data = value;
                Checksum = CalcChecksum();
            }
        }
        public virtual byte this[int index]
        {
            get
            {
                return Data[index];
            }
            set
            {
                Data[index] = value;
                Checksum = CalcChecksum();
            }
        }

        public virtual ushort Checksum { get; protected set; }

        public virtual int DataSize { get { return Data.Length; } }

        protected virtual void LoadBlock(Stream input, int blockSize, int footerSize)
        {
            // is there enough data?
            int dataSize = blockSize - footerSize;
            if (input.Length - input.Position < blockSize)
                throw new ArgumentOutOfRangeException("dataSize", blockSize, "Not enough bytes to read block (required: at least " + footerSize + " bytes).");

            // read the data
            data = new byte[dataSize];
            input.Read(data, 0, dataSize);

            LoadFooter(input);
        }
        protected abstract void LoadFooter(Stream input);
        protected virtual void CheckChecksum()
        {
            ushort correctChecksum = CalcChecksum();
            if (Checksum != correctChecksum)
                throw new InvalidDataException("Checksum 0x" + Checksum.ToString("X4") + " is incorrect (expected: 0x" + correctChecksum.ToString("X4") + ").");
        }
        protected virtual ushort CalcChecksum()
        {
            return CRC16CCITT.Calculate(data);
        }
        public virtual void FixChecksum()
        {
            Checksum = CalcChecksum();
        }

        public virtual MemoryStream ToData()
        {
            MemoryStream output = new MemoryStream();

            output.Write(Data, 0, DataSize);
            WriteFooter(output);

            return output;
        }
        public virtual void WriteTo(Stream output)
        {
            ToData().WriteTo(output);
        }
        protected abstract void WriteFooter(Stream output);

        protected virtual void Align(Stream stream)
        {
            long target = (stream.Position + 3) & 4;
            while (target > stream.Position)
                stream.WriteByte(0);
        }
    }
}
