﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Text;
using System.Windows.Forms;
using PWTEditor.SaveFile;
using PWTEditor.SaveFile.Tournaments;
using PWTEditor.Tournaments;

namespace PWTEditor
{
    public partial class SaveOverview : Form
    {
        private ValueNames ValueNames = new ValueNames();
        private static readonly int tournamentCount = 3;
        public int[] itemIndexTable { get; set; }
        string formTitle; // used as storage at runtime

        byte[] saveFile;
        NoSizeBlock checksumListBlock;
        GenericBlock indexBlock;

        BindingList<Tournament> tournaments = new BindingList<Tournament>();
        TournamentBlock[] tournamentBlocks = new TournamentBlock[tournamentCount];

        // properties
        bool SaveLoaded
        {
            get
            {
                return saveFile != null;
            }
        }
        int TournamentIndex
        {
            get
            {
                return tournamentListBox.SelectedIndex;
            }
        }
        bool TournamentSelected
        {
            get
            {
                return TournamentIndex >= 0 && TournamentIndex < tournamentCount;
            }
        }
        Tournament CurrentTournament
        {
            get
            {
                if (TournamentSelected)
                    return tournaments[TournamentIndex];
                else
                    return new Tournament();
            }
            set
            {
                if (TournamentSelected)
                    tournaments[TournamentIndex] = value;
            }
        }
        TournamentBlock CurrentTournamentBlock
        {
            get
            {
                if (TournamentSelected)
                    return tournamentBlocks[TournamentIndex];
                else
                    return null;
            }
            set
            {
                if (TournamentSelected)
                    tournamentBlocks[TournamentIndex] = value;
            }
        }

        public SaveOverview()
        {
            InitializeComponent();

            // append version number to title
            this.Text += ' ' + Program.Version;
            formTitle = this.Text;

            // load type names
            monotypeComboBox.Items.AddRange(ValueNames.Types);
            // load Pokémon names, but exclude everything before Bulbasaur and after Genesect
            LoadNamesSorted(bannedPokemonCheckedListBox, ValueNames.Pokemon, 1, 649, false);
            // load item names, but exclude no-item and everything after Casteliacone
            itemIndexTable = LoadNamesSorted(bannedItemsCheckedListBox, ValueNames.Items, 1, 591, true);
            // load move names
            bannedMoveComboBox.Items.AddRange(ValueNames.Moves);

            // autosize description and rules text boxes
            descriptionTextBox.Height = 7 + descriptionTextBox.Font.Height * 2;
            rulesTextBox.Height = 7 + rulesTextBox.Font.Height * 3;
        }
        private int[] LoadNamesSorted(CheckedListBox comboBox, string[] names, int start, int count, bool sort)
        {
            string[] subNames = new string[count];
            Array.Copy(names, start, subNames, 0, count);

            List<string> subNamesList = new List<string>(subNames);
            List<string> sortedSubNames = new List<string>(subNames);

            int[] indexTable = null;
            if (sort)
            {
                sortedSubNames.Sort(SortNames);

                indexTable = new int[count];
                for (int i = 0; i < subNames.Length; i++)
                {
                    indexTable[i] = subNamesList.IndexOf(sortedSubNames[i]);
                }
            }

            comboBox.Items.AddRange(sortedSubNames.ToArray());
            return indexTable;
        }
        private int SortNames(string a, string b)
        {
            if (a == b)
                return 0;
            if (a.Length == 0)
                return -1; // a -> b
            if (b.Length == 0)
                return 1; // b -> a
            if (Char.IsLetterOrDigit(a[0]))
            {
                if (Char.IsLetterOrDigit(b[0]))
                {
                    if (a[0] != b[0])
                        return a.CompareTo(b); // both are different letter/digit
                    else
                        return SortNames(a.Substring(1), b.Substring(1)); // same letter/digit, so move on to next character
                }
                else
                    return -1; // a -> b, as a is a letter/digit and b is not
            }
            else
            {
                if (Char.IsLetterOrDigit(b[0]))
                    return 1; // b -> a, as b is a letter/digit and a is not
                else
                {
                    if (a[0] != b[0])
                        return a.CompareTo(b); // neither are letters or digits, but different
                    else
                        return SortNames(a.Substring(1), b.Substring(1)); // same nonletter/nondigit, so move on to next character
                }
            }
        }

        private void loadSAVToolStripMenuItem_Click(object sender, EventArgs e)
        {
            openSavDialog.ShowDialog();
        }
        private void saveSAVToolStripMenuItem_Click(object sender, EventArgs e)
        {
            saveSavDialog.ShowDialog();
        }
        private void importPWTToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (SaveLoaded && TournamentSelected)
                openPWTDialog.ShowDialog();
        }
        private void exportPWTToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (SaveLoaded && TournamentSelected && !CurrentTournament.IsEmpty)
            {
                // set a default name for the tournament file
                savePWTDialog.FileName = CurrentTournament.TournamentName;
                savePWTDialog.ShowDialog();
            }
        }
        private void deleteToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (SaveLoaded && TournamentSelected && !CurrentTournament.IsEmpty)
            {
                CurrentTournament = new Tournament();
                CurrentTournamentBlock = null;
            }
        }
        private void loadRawDataToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (SaveLoaded && TournamentSelected)
                openDataDialog.ShowDialog();
        }
        private void dumpRawDataToolStripMenuItem_Click(object sender, EventArgs e)
        {
            if (SaveLoaded && TournamentSelected && !CurrentTournament.IsEmpty)
            {
                // set a default name for the tournament file
                saveDataDialog.FileName = savePWTDialog.FileName = CurrentTournament.TournamentName;
                saveDataDialog.ShowDialog();
            }
        }

        private void openSavDialog_FileOk(object sender, CancelEventArgs e)
        {
            // copy loaded file name to the save dialog
            saveSavDialog.FileName = openSavDialog.FileName;
            this.Text = formTitle + " - " + Path.GetFileName(openSavDialog.FileName);

            // try to cache the save file
            if (!LoadFile(openSavDialog.FileName, "save file", ref saveFile, 0x80000))
            {
                e.Cancel = true;
                return;
            }

            // make save file a stream for easier reading
            MemoryStream saveStream = new MemoryStream(saveFile);

            // try to load the checksum list block; if this fails, the save is corrupt
            NoSizeBlock checksumListBlock = LoadChecksumListBlock(saveStream);
            if (checksumListBlock == null)
            {
                e.Cancel = true;
                return;
            }

            // load the block index block; if this fails, the save is corrupt
            GenericBlock indexBlock = LoadIndexBlock(saveStream);
            if (indexBlock == null)
            {
                e.Cancel = true;
                return;
            }

            // at this point, the save has been loaded correctly
            this.checksumListBlock = checksumListBlock;
            this.indexBlock = indexBlock;

            // load the tournaments
            LoadTournaments(saveStream);

            PostLoad();
        }
        private void saveSavDialog_FileOk(object sender, CancelEventArgs e)
        {
            // copy saved file name to the open dialog
            openSavDialog.FileName = saveSavDialog.FileName;
            this.Text = formTitle + " - " + Path.GetFileName(saveSavDialog.FileName);

            // make save file a stream for easier writing
            MemoryStream saveStream = new MemoryStream(saveFile);

            // write the tournaments to the save
            WriteTournaments(saveStream);

            // write save file to file
            saveFile = saveStream.ToArray();
            if (!WriteFile(saveSavDialog.FileName, "save file", saveFile))
            {
                e.Cancel = true;
                return;
            }
        }
        private void openPWTDialog_FileOk(object sender, CancelEventArgs e)
        {
            // copy loaded file name to the save dialog
            savePWTDialog.FileName = openPWTDialog.FileName;

            // load the tournament
            byte[] tournamentBlockBytes = null;
            if (!LoadFile(openPWTDialog.FileName, "tournament", ref tournamentBlockBytes, 0x1314))
            {
                e.Cancel = true;
                return;
            }

            // import the tournament
            MemoryStream tournamentBlockStream = new MemoryStream(tournamentBlockBytes);
            Tournament tournament = null;
            try
            {
                TournamentBlock tournamentBlock = new TournamentBlock(tournamentBlockStream);
                tournament = tournamentBlock.TournamentChecksumBlock.TournamentData.Tournament;

                CurrentTournament = tournament;
                CurrentTournamentBlock = tournamentBlock;
            }
            catch
            {
                MessageBox.Show("The tournament could not be imported!\nIt may be corrupted.", "Error");
                e.Cancel = true;
                return;
            }
        }
        private void savePWTDialog_FileOk(object sender, CancelEventArgs e)
        {
            // copy saved file name to the open dialog
            openPWTDialog.FileName = savePWTDialog.FileName;

            // export the tournament
            byte[] tournamentBlockBytes = CurrentTournamentBlock.ToData().ToArray();

            // save the tournament
            if (!WriteFile(savePWTDialog.FileName, "tournament", tournamentBlockBytes))
            {
                e.Cancel = true;
                return;
            }
        }
        private void openDataDialog_FileOk(object sender, CancelEventArgs e)
        {
            // copy loaded file name to the save dialog
            saveDataDialog.FileName = openDataDialog.FileName;

            // load the tournament data
            byte[] tournamentBytes = null;
            if (!LoadFile(openDataDialog.FileName, "data file", ref tournamentBytes, 0x1210))
            {
                e.Cancel = true;
                return;
            }

            // import the tournament data
            CurrentTournament = new Tournament(tournamentBytes);
            if (CurrentTournamentBlock == null)
            {
                // create a new tournament block
                TournamentData tournamentData = new TournamentData(CurrentTournament);
                TournamentChecksumBlock tournamentChecksumBlock = new TournamentChecksumBlock(tournamentData);
                CurrentTournamentBlock = new TournamentBlock(tournamentChecksumBlock);
                CurrentTournamentBlock.SaveCount = 1;
            }
        }
        private void saveDataDialog_FileOk(object sender, CancelEventArgs e)
        {
            // copy saved file name to the open dialog
            openDataDialog.FileName = saveDataDialog.FileName;

            // save the tournament data
            if (!WriteFile(saveDataDialog.FileName, "data file", CurrentTournament.Data))
            {
                e.Cancel = true;
                return;
            }
        }

        private void ShowTournamentLoadErrors(bool[] successFlags)
        {
            List<int> errors = new List<int>();
            for (int i = 0; i < successFlags.Length; i++)
            {
                if (!successFlags[i])
                    errors.Add(i + 1);
            }
            if (errors.Count > 0)
            {
                // generate error message
                string message = " Tournament" + (errors.Count != 1 ? "s" : "");
                for (int i = 0; i < errors.Count; i++)
                {
                    if (i >= 1)
                    {
                        if (i == errors.Count - 1)
                            message += " and ";
                        else
                            message += ", ";
                    }

                    message += " " + errors[i];
                }
                message += errors.Count == 1 ? " is" : " are";
                message += " corrupted and ";
                message += errors.Count == 1 ? "has" : "have";
                message += " been discarded.";

                MessageBox.Show(message, "Warning");
            }
        }
        private void PostLoad()
        {
            bool enable = SaveLoaded;
            tournamentsGroupBox.Enabled = enable;

            tournamentListBox.DisplayMember = "TournamentName";
            tournamentListBox.DataSource = tournaments;
        }
        private void TournamentEnable(bool enable)
        {
            bool selected = TournamentSelected && enable;
            tournamentToolStripMenuItem.Enabled = selected;
            dataTabControl.Enabled = selected;
        }

        private bool LoadFile(string path, string fileTypeName, ref byte[] file, int size)
        {
            bool fileLoaded = false;
            while (!fileLoaded)
            {
                // try loading the file
                try
                {
                    file = new byte[size];
                    using (FileStream input = new FileStream(path, FileMode.Open))
                    {
                        input.Read(file, 0, size);
                    }
                    fileLoaded = true;
                }
                catch
                {
                    fileLoaded = false;
                }

                if (!fileLoaded)
                {
                    // ask user to try again
                    if (MessageBox.Show("The " + fileTypeName + " could not be accessed.\nIt may be in use by a different program.", "Error", MessageBoxButtons.RetryCancel) == DialogResult.Cancel)
                        return false;
                }
            }

            return true;
        }
        private bool WriteFile(string path, string fileTypeName, byte[] file)
        {
            bool fileSaved = false;
            while (!fileSaved)
            {

                try
                {
                    using (FileStream output = new FileStream(path, FileMode.Create))
                    {
                        output.Write(file, 0, file.Length);
                    }
                    fileSaved = true;
                }
                catch
                {
                    fileSaved = false;
                }

                if (!fileSaved)
                {
                    // ask user to try again
                    if (MessageBox.Show("The " + fileTypeName + " could not be written.\nIt may be in use by a different program.", "Error", MessageBoxButtons.RetryCancel) == DialogResult.Cancel)
                        return false;
                }
            }

            return true;
        }
        private NoSizeBlock LoadChecksumListBlock(Stream saveStream)
        {
            try
            {
                saveStream.Position = 0x25F00;
                return new NoSizeBlock(saveStream, 0xA4);
            }
            catch
            {
                try
                {
                    saveStream.Position = 0x4BF00;
                    return new NoSizeBlock(saveStream, 0xA4);
                }
                catch
                {
                    MessageBox.Show("The checksum list in the save file could not be loaded.\nThe save file may be corrupted.", "Error");
                    return null;
                }
            }
        }
        private GenericBlock LoadIndexBlock(Stream saveStream)
        {
            try
            {
                saveStream.Position = 0x19400;
                return new GenericBlock(saveStream, 0xB4);
            }
            catch
            {
                try
                {
                    saveStream.Position = 0x3F400;
                    return new GenericBlock(saveStream, 0xB4);
                }
                catch
                {
                    MessageBox.Show("The block index in the save file could not be loaded.\nThe save file may be corrupted.", "Error");
                    return null;
                }
            }
        }
        private void LoadTournaments(Stream saveStream)
        {
            // clear existing tournaments
            tournamentBlocks = new TournamentBlock[tournamentCount];
            tournaments = new BindingList<Tournament>();

            bool[] successFlags = new bool[tournamentCount];
            int tournamentNumber = 0;
            while (tournamentNumber < tournamentCount)
            {
                try
                {
                    // two while loops to eliminate unnecessary try-catch overhead
                    while (tournamentNumber < tournamentCount)
                    {
                        // does the block exist?
                        if (BitConverter.ToUInt16(indexBlock.Data, 0x52 + 2 * tournamentNumber) != 0xC21E)
                        {
                            // set an empty tournament
                            tournaments.Add(new Tournament());
                            // block doesn't exist, so just treat it as valid
                            successFlags[tournamentNumber++] = true;
                            continue;
                        }

                        // load tournament
                        saveStream.Position = 0x7A000 + 0x1400 * tournamentNumber;
                        TournamentBlock tournamentBlock = new TournamentBlock(saveStream);
                        Tournament tournament = tournamentBlock.TournamentChecksumBlock.TournamentData.Tournament;

                        // tournament loaded successfully; store it and move on to the next tournament
                        tournamentBlocks[tournamentNumber] = tournamentBlock;
                        tournaments.Add(tournament);
                        successFlags[tournamentNumber++] = true;
                    }
                }
                catch
                {
                    // set an empty tournament
                    tournaments.Add(new Tournament());
                    // move on to the next tournament
                    tournamentNumber++;
                }
            }

            ShowTournamentLoadErrors(successFlags);
        }
        private void WriteTournaments(Stream saveStream)
        {
            for (int i = 0; i < tournamentCount; i++)
            {
                if (tournamentBlocks[i] == null)
                {
                    // tournament doesn't exist, so clear it from the index
                    BitConverter.GetBytes((ushort)0x0000).CopyTo(indexBlock.Data, 0x52 + i * 2);
                    BitConverter.GetBytes(0).CopyTo(indexBlock.Data, 0xA0 + i * 4);
                }
                else
                {
                    // unpack tournament block
                    TournamentChecksumBlock tournamentChecksumBlock = tournamentBlocks[i].TournamentChecksumBlock;
                    TournamentData tournamentData = tournamentChecksumBlock.TournamentData;

                    // replace the tournament
                    tournamentData.Tournament = tournaments[i];

                    // pack tournament block
                    tournamentChecksumBlock.TournamentData = tournamentData;
                    tournamentBlocks[i].TournamentChecksumBlock = tournamentChecksumBlock;

                    // write tournament block
                    saveStream.Position = 0x7A000 + 0x1400 * i;
                    tournamentBlocks[i].WriteTo(saveStream);

                    // set it in index
                    BitConverter.GetBytes((ushort)0xC21E).CopyTo(indexBlock.Data, 0x52 + i * 2);
                    BitConverter.GetBytes(tournamentBlocks[i].SaveCount).CopyTo(indexBlock.Data, 0xA0 + i * 4);
                }
            }
            indexBlock.FixChecksum();

            // write the index block to the save
            saveStream.Position = 0x19400;
            indexBlock.WriteTo(saveStream);
            saveStream.Position = 0x3F400;
            indexBlock.WriteTo(saveStream);

            // update the index block checksum in the checksum list
            checksumListBlock[0x36] = (byte)indexBlock.Checksum;
            checksumListBlock[0x37] = (byte)(indexBlock.Checksum >> 8);

            // write the checksum list to the save
            saveStream.Position = 0x25F00;
            checksumListBlock.WriteTo(saveStream);
            saveStream.Position = 0x4BF00;
            checksumListBlock.WriteTo(saveStream);
        }

        private void BindTournament(object sender, EventArgs e)
        {
            // bind tournament name text box
            nameTextBox.DataBindings.Clear();
            nameTextBox.DataBindings.Add("Text", CurrentTournament, "TournamentName");
            // bind tournament description text box
            descriptionTextBox.DataBindings.Clear();
            descriptionTextBox.DataBindings.Add("Text", CurrentTournament, "TournamentDescription");
            // bind tournament rules text box
            rulesTextBox.DataBindings.Clear();
            rulesTextBox.DataBindings.Add("Text", CurrentTournament, "TournamentRules");

            // bind battle type combo box
            battleTypeComboBox.DataBindings.Clear();
            battleTypeComboBox.DataBindings.Add("SelectedIndex", CurrentTournament, "BattleType");
            // bind level style controls
            levelStyleComboBox.DataBindings.Clear();
            levelStyleComboBox.DataBindings.Add("SelectedIndex", CurrentTournament, "LevelStyle");
            levelNumericUpDown.DataBindings.Clear();
            levelNumericUpDown.DataBindings.Add("Value", CurrentTournament, "AllowedLevel");
            partyLevelNumericUpDown.DataBindings.Clear();
            partyLevelNumericUpDown.DataBindings.Add("Value", CurrentTournament, "TotalLevel");
            // bind Pokémon count numeric up/downs
            minimumPokemonNumericUpDown.DataBindings.Clear();
            minimumPokemonNumericUpDown.DataBindings.Add("Value", CurrentTournament, "MinimumPokemon");
            maximumPokemonNumericUpDown.DataBindings.Clear();
            maximumPokemonNumericUpDown.DataBindings.Add("Value", CurrentTournament, "MaximumPokemon");
            // bind options controls
            duplicatePokemonCheckBox.DataBindings.Clear();
            duplicatePokemonCheckBox.DataBindings.Add("Checked", CurrentTournament, "DuplicatePokemon");
            duplicateItemsCheckBox.DataBindings.Clear();
            duplicateItemsCheckBox.DataBindings.Add("Checked", CurrentTournament, "DuplicateItems");
            worldTournamentCheckBox.DataBindings.Clear();
            worldTournamentCheckBox.DataBindings.Add("Checked", CurrentTournament, "WorldTournament");
            monotypeComboBox.DataBindings.Clear();
            monotypeComboBox.DataBindings.Add("SelectedIndex", CurrentTournament, "Monotype");

            // pseudo-bind banlists
            CheckedListBoxBind(bannedPokemonCheckedListBox, "BannedPokemon");
            CheckedListBoxBind(bannedItemsCheckedListBox, "BannedItems", "itemIndexTable");
            bannedMoveComboBox.DataBindings.Clear();
            bannedMoveComboBox.DataBindings.Add("SelectedIndex", CurrentTournament, "BannedMove");

            if (CurrentTournament.IsEmpty)
            {
                TournamentEnable(false);
                tournamentToolStripMenuItem.Enabled = true;
            }
            else
                TournamentEnable(true);
        }
        private void EnterPressUpdate(object sender, KeyEventArgs e)
        {
            if (e.KeyCode == Keys.Enter)
            {
                UpdateValue(sender);
            }
        }
        private void UpdateValue(object sender, EventArgs e = null)
        {
            ControlBindingsCollection dataBindings = ((Control)sender).DataBindings;
            if (dataBindings.Count > 0)
                dataBindings[0].WriteValue();
        }

        private void CheckedListBoxBind(CheckedListBox checkedListBox, string variable, string indexTableName = null)
        {
            // load bool array from tournament
            bool[] checkedItems = (bool[])typeof(Tournament).GetProperty(variable).GetValue(CurrentTournament, null);
            // if index table exists, load index table
            int[] indexTable = null;
            if (indexTableName != null)
                indexTable = (int[])this.GetType().GetProperty(indexTableName).GetValue(this, null);

            // unbind
            checkedListBox.ItemCheck -= CheckedListBoxUpdate;

            // check all necessary items
            if (indexTable != null)
            {
                for (int i = 0; i < checkedItems.Length && i < checkedListBox.Items.Count; i++)
                    checkedListBox.SetItemChecked(i, checkedItems[indexTable[i]]);
            }
            else
            {
                for (int i = 0; i < checkedItems.Length && i < checkedListBox.Items.Count; i++)
                    checkedListBox.SetItemChecked(i, checkedItems[i]);
            }

            // rebind
            checkedListBox.ItemCheck += CheckedListBoxUpdate;
            checkedListBox.Tag = variable + (indexTable != null ? "," + indexTableName : "" );
        }
        private void CheckedListBoxUpdate(object sender, ItemCheckEventArgs e)
        {
            CheckedListBox checkedListBox = (CheckedListBox)sender;
            string variable;
            int[] indexTable = null;
            string tag = (string)checkedListBox.Tag;
            int commaPos = tag.IndexOf(',');
            if (commaPos == -1)
            {
                variable = tag;
            }
            else
            {
                variable = tag.Substring(0, commaPos);
                indexTable = (int[])this.GetType().GetProperty(tag.Substring(commaPos + 1)).GetValue(this, null);
            }

            bool[] checkedItems = new bool[checkedListBox.Items.Count];

            if (indexTable == null)
            {
                foreach (int checkedItemIndex in checkedListBox.CheckedIndices)
                    checkedItems[checkedItemIndex] = true;
                checkedItems[e.Index] = e.NewValue != CheckState.Unchecked;
            }
            else
            {
                foreach (int checkedItemIndex in checkedListBox.CheckedIndices)
                    checkedItems[indexTable[checkedItemIndex]] = true;
                checkedItems[indexTable[e.Index]] = e.NewValue != CheckState.Unchecked;
            }

            typeof(Tournament).GetProperty(variable).SetValue(CurrentTournament, checkedItems, null);
        }

        private void simpleComboBox_Leave(object sender, EventArgs e)
        {
            ComboBox comboBox = (ComboBox)sender;
            if (!comboBox.Items.Contains(comboBox.Text))
                comboBox.SelectedIndex = 0;
        }
    }
}