﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Text;
using PWTEditor.SaveFile.Tournaments;

namespace PWTEditor.Tournaments
{
    public class Tournament : INotifyPropertyChanged
    {
        public event PropertyChangedEventHandler PropertyChanged;
        private UnicodeEncoding unicode = new UnicodeEncoding(false, false);
        private byte[] data;
        public byte[] Data
        {
            get
            {
                if (data != null)
                    return data;
                else
                    return new byte[1210];
            }
            set
            {
                data = value;
            }
        }

        public Boolean IsEmpty { get; protected set; }

        public Tournament(TournamentData tournamentData) : this(tournamentData.Data) { }
        public Tournament(byte[] data)
        {
            this.Data = data;
            this.IsEmpty = false;
        }
        public Tournament()
        {
            this.Data = new byte[0x1210];
            this.IsEmpty = true;
        }

        public Tournament Clone()
        {
            return new Tournament(Data);
        }

        private string GetString(int offset, int length)
        {
            return unicode
                .GetString(Data, offset, length * 2)
                .Replace("\uFFFE", "\r\n")
                .TrimEnd('\uFFFF');
        }
        private void SetString(string input, int offset, int length)
        {
            byte[] bytes = unicode.GetBytes(input
                    .Replace("\r", "")
                    .Replace("\n", "\uFFFE")
                    .PadRight(length, '\uFFFF')
                .ToCharArray(), 0, length);
            Array.Copy(bytes, 0, Data, offset, length * 2);
        }
        private bool[] GetBits(int offset, int start, int length)
        {
            bool[] setBits = new bool[length];

            
            int address;
            int bitMask;
            for (int i = 0, index = start; i < length; i++, index++)
            {
                address = offset + (index >> 3);
                bitMask = 1 << (index & 7);
                if ((Data[address] & bitMask) != 0)
                    setBits[i] = true;
            }

            return setBits;
        }
        private void SetBits(bool[] input, int offset, int start, int length)
        {
            byte[] bytes = new byte[(length + 7) >> 3];

            // note: skips everything before Bulbasaur and after Genesect
            int address;
            int bitMask;
            for (int i = 0, index = start; i < length && i < input.Length; i++, index++)
            {
                if (input[i])
                {
                    address = index >> 3;
                    bitMask = 1 << (index & 7);
                    bytes[address] |= (byte)bitMask;
                }
            }

            Array.Copy(bytes, 0, Data, offset, bytes.Length);
        }
        private void SetUInt16(int input, int offset)
        {
            Data[offset] = (byte)input;
            Data[offset + 1] = (byte)(input >> 8);
        }

        [Bindable(true)]
        public string TournamentName
        {
            get
            {
                if (IsEmpty)
                    return "Empty";
                else
                    return GetString(0xA, 37);
            }
            set
            {
                SetString(value, 0xA, 37);
                PropertyChanged(this, new PropertyChangedEventArgs("TournamentName"));
                IsEmpty = false;
            }
        }
        public string TournamentDescription
        {
            get
            {
                if (IsEmpty)
                    return "No description";
                else
                    return GetString(0x54, 75);
            }
            set
            {
                SetString(value, 0x54, 75);
                IsEmpty = false;
            }
        }
        public string TournamentRules
        {
            get
            {
                if (IsEmpty)
                    return "No rules";
                else
                    return GetString(0xEA, 111);
            }
            set
            {
                SetString(value, 0xEA, 111);
                IsEmpty = false;
            }
        }

        public int BattleType
        {
            get
            {
                return Data[0x26F] & 3;
            }
            set
            {
                Data[0x26F] = (byte)(value & 3);
                IsEmpty = false;
            }
        }
        public int LevelStyle
        {
            get
            {
                int battleStyle = Data[0x1CB];
                if (battleStyle <= 5)
                    return battleStyle;
                else
                    return 0;
            }
            set
            {
                if (value <= 5)
                    Data[0x1CB] = (byte)value;
                else
                    Data[0x1CB] = 0;

                IsEmpty = false;
            }
        }
        public decimal AllowedLevel
        {
            get
            {
                int allowedLevel = Data[0x1CA];
                if (allowedLevel <= 100)
                    return allowedLevel;
                else
                    return 100;
            }
            set
            {
                if (value <= 100)
                    Data[0x1CA] = (byte)value;
                else
                    Data[0x1CA] = 100;

                IsEmpty = false;
            }
        }
        public decimal TotalLevel
        {
            get
            {
                int totalLevel = BitConverter.ToUInt16(Data, 0x1CC);
                if (totalLevel > 0 && totalLevel <= 600)
                    return totalLevel;
                else
                    return 600;
            }
            set
            {
                if (value <= 600)
                    SetUInt16((int)value, 0x1CC);
                else // if total level >= 600, just disable total level altogether
                    SetUInt16(0, 0x1CC);

                IsEmpty = false;
            }
        }

        [Bindable(true)]
        public decimal MinimumPokemon
        {
            get
            {
                int minimumPokemon = Data[0x1C8];
                if (minimumPokemon == 0)
                    return 1;
                else if (minimumPokemon <= 6)
                    return minimumPokemon;
                else
                    return 6;
            }
            set
            {
                if (value <= 6)
                    Data[0x1C8] = (byte)value;
                else
                    Data[0x1C8] = 6;

                // if min > max, set max to min
                int minimumPokemon = Data[0x1C8];
                int maximumPokemon = Data[0x1C9];
                if (minimumPokemon > maximumPokemon)
                {
                    Data[0x1C9] = (byte)minimumPokemon;
                    PropertyChanged(this, new PropertyChangedEventArgs("MaximumPokemon"));
                }

                IsEmpty = false;
            }
        }
        [Bindable(true)]
        public decimal MaximumPokemon
        {
            get
            {
                int minimumPokemon = Data[0x1C8];
                int maximumPokemon = Data[0x1C9];
                if (maximumPokemon == 0)
                    return 1;
                if (maximumPokemon <= 6)
                {
                    // if min > max, set max to min (so they are both set to the highest value)
                    if (maximumPokemon >= minimumPokemon)
                        return maximumPokemon;
                    else
                        return minimumPokemon;
                }
                else
                    return 6;
            }
            set
            {
                if (value <= 6)
                    Data[0x1C9] = (byte)value;
                else
                    Data[0x1C9] = 6;

                // if min > max, set min to max
                int minimumPokemon = Data[0x1C8];
                int maximumPokemon = Data[0x1C9];
                if (minimumPokemon > maximumPokemon)
                {
                    Data[0x1C8] = (byte)maximumPokemon;
                    PropertyChanged(this, new PropertyChangedEventArgs("MinimumPokemon"));
                }

                IsEmpty = false;
            }
        }

        public bool DuplicatePokemon
        {
            get
            {
                return Data[0x1CE] != 0;
            }
            set
            {
                Data[0x1CE] = (byte)(value ? 1 : 0);
                IsEmpty = false;
            }
        }
        public bool DuplicateItems
        {
            get
            {
                return Data[0x1CF] != 0;
            }
            set
            {
                Data[0x1CF] = (byte)(value ? 1 : 0);
                IsEmpty = false;
            }
        }
        public bool WorldTournament
        {
            get
            {
                return Data[0x270] != 0;
            }
            set
            {
                Data[0x270] = (byte)(value ? 1 : 0);
                IsEmpty = false;
            }
        }
        public int Monotype
        {
            get
            {
                int monotype = Data[0x271];
                if (!IsEmpty && monotype <= 17)
                    return monotype;
                else
                    return 17;
            }
            set
            {
                if (value >= 0 && value <= 17)
                    Data[0x271] = (byte)value;
                else
                    Data[0x271] = 17;

                IsEmpty = false;
            }
        }

        public bool[] BannedPokemon
        {
            // note: ignores/deletes everything before Bulbasaur and after Genesect
            get
            {
                return GetBits(0x1D0, 1, 649);
            }
            set
            {
                SetBits(value, 0x1D0, 1, 649);
                IsEmpty = false;
            }
        }
        public bool[] BannedItems
        {
            // note: ignores/deletes no-item and everything after Casteliacone
            get
            {
                return GetBits(0x222, 1, 591);
            }
            set
            {
                SetBits(value, 0x222, 1, 591);
                IsEmpty = false;
            }
        }
        public int BannedMove
        {
            get
            {
                int bannedMove = BitConverter.ToUInt16(Data, 0x272);
                if (bannedMove <= 559)
                    return bannedMove;
                else
                    return 0;
            }
            set
            {
                if (value >= 0 && value <= 559)
                    SetUInt16(value, 0x272);
                else
                    SetUInt16(0, 0x272);

                IsEmpty = false;
            }
        }
    }
}
